@@ -25,13 +25,55 @@
 #include <gst/audio/audio.h>
 #include <gst/video/video.h>
 
+/**
+ * GstFFMpegCompliance:
+ * @GST_FFMPEG_VERY_STRICT: Strictly conform to an older
+ * more strict version of the spec or reference software
+ * @GST_FFMPEG_STRICT: Strictly conform to all the things
+ * in the spec no matter what consequences.
+ * @GST_FFMPEG_NORMAL:
+ * @GST_FFMPEG_UNOFFICIAL: Allow unofficial extensions
+ * @GST_FFMPEG_EXPERIMENTAL: Allow nonstandardized
+ * experimental things.
+ *
+ * This setting instructs libav on how strictly it should follow the
+ * associated standard.
+ *
+ * From avcodec.h:
+ * Setting this to STRICT or higher means the encoder and decoder will
+ * generally do stupid things, whereas setting it to unofficial or lower
+ * will mean the encoder might produce output that is not supported by all
+ * spec-compliant decoders. Decoders don't differentiate between normal,
+ * unofficial and experimental (that is, they always try to decode things
+ * when they can) unless they are explicitly asked to behave stupidly
+ * (=strictly conform to the specs)
+ */
+typedef enum {
+  GST_FFMPEG_VERY_STRICT = FF_COMPLIANCE_VERY_STRICT,
+  GST_FFMPEG_STRICT = FF_COMPLIANCE_STRICT,
+  GST_FFMPEG_NORMAL = FF_COMPLIANCE_NORMAL,
+  GST_FFMPEG_UNOFFICIAL = FF_COMPLIANCE_UNOFFICIAL,
+  GST_FFMPEG_EXPERIMENTAL = FF_COMPLIANCE_EXPERIMENTAL,
+} GstFFMpegCompliance;
+
+/*
+ * _compliance_get_type () Returns an enum type that can be
+ * used as a property to indicate desired FFMpeg adherence to
+ * an associated specification
+ */
+
+GType
+gst_ffmpeg_compliance_get_type (void);
+#define GST_TYPE_FFMPEG_COMPLIANCE (gst_ffmpeg_compliance_get_type ())
+#define FFMPEG_DEFAULT_COMPLIANCE GST_FFMPEG_NORMAL
+
 /*
  * _codecid_to_caps () gets the GstCaps that belongs to
  * a certain CodecID for a pad with compressed data.
  */
 
 GstCaps *
-gst_ffmpeg_codecid_to_caps   (enum CodecID    codec_id,
+gst_ffmpeg_codecid_to_caps   (enum AVCodecID    codec_id,
                               AVCodecContext *context,
                               gboolean        encode);
 
@@ -42,12 +84,12 @@ gst_ffmpeg_codecid_to_caps   (enum CodecID    codec_id,
 
 GstCaps *
 gst_ffmpeg_codectype_to_audio_caps (AVCodecContext *context, 
-                              enum CodecID codec_id,
+                              enum AVCodecID codec_id,
 				    gboolean encode,
 				    AVCodec *codec);
 GstCaps *
 gst_ffmpeg_codectype_to_video_caps (AVCodecContext *context, 
-                              enum CodecID codec_id,
+                              enum AVCodecID codec_id,
 				    gboolean encode,
 				    AVCodec *codec);
 
@@ -57,7 +99,7 @@ gst_ffmpeg_codectype_to_video_caps (AVCodecContext *context,
  * context and a codecID.
  */
 
-enum CodecID
+enum AVCodecID
 gst_ffmpeg_caps_to_codecid (const GstCaps  *caps,
                             AVCodecContext *context);
 
@@ -67,7 +109,7 @@ gst_ffmpeg_caps_to_codecid (const GstCaps  *caps,
  */
 
 void
-gst_ffmpeg_caps_with_codecid (enum CodecID    codec_id,
+gst_ffmpeg_caps_with_codecid (enum AVCodecID    codec_id,
                               enum AVMediaType  codec_type,
                               const GstCaps  *caps,
                               AVCodecContext *context);
@@ -113,8 +155,8 @@ gst_ffmpeg_formatid_to_caps (const gchar *format_name);
 
 gboolean
 gst_ffmpeg_formatid_get_codecids (const gchar *format_name,
-                                  enum CodecID ** video_codec_list,
-                                  enum CodecID ** audio_codec_list,
+                                  enum AVCodecID ** video_codec_list,
+                                  enum AVCodecID ** audio_codec_list,
 				  AVOutputFormat * plugin);
 
 
