@@ -815,7 +815,7 @@ gst_ffmpegviddec_get_buffer2 (AVCodecContext * context, AVFrame * picture,
   /* Fill avpicture */
   if (!gst_video_frame_map (&dframe->vframe, &ffmpegdec->pool_info,
           dframe->buffer, GST_MAP_READWRITE))
-    goto invalid_frame;
+    goto map_failed;
   dframe->mapped = TRUE;
 
   for (c = 0; c < AV_NUM_DATA_POINTERS; c++) {
@@ -850,30 +850,13 @@ gst_ffmpegviddec_get_buffer2 (AVCodecContext * context, AVFrame * picture,
 
   return 0;
 
-  /* fallbacks */
 no_dr:
-  {
-    GST_LOG_OBJECT (ffmpegdec, "direct rendering disabled, fallback alloc");
-    goto fallback;
-  }
-alloc_failed:
-  {
-    /* alloc default buffer when we can't get one from downstream */
-    GST_LOG_OBJECT (ffmpegdec, "alloc failed, fallback alloc");
-    goto fallback;
-  }
-invalid_frame:
-  {
-    /* alloc default buffer when we can't get one from downstream */
-    GST_LOG_OBJECT (ffmpegdec, "failed to map frame, fallback alloc");
-    gst_buffer_replace (&dframe->buffer, NULL);
-    goto fallback;
-  }
-fallback:
   {
     int c;
     int ret = avcodec_default_get_buffer2 (context, picture, flags);
 
+    GST_LOG_OBJECT (ffmpegdec, "direct rendering disabled, fallback alloc");
+
     for (c = 0; c < AV_NUM_DATA_POINTERS; c++) {
       ffmpegdec->stride[c] = picture->linesize[c];
     }
@@ -893,6 +876,21 @@ fallback:
 
     return ret;
   }
+alloc_failed:
+  {
+    GST_ELEMENT_ERROR (ffmpegdec, RESOURCE, FAILED,
+        ("Unable to allocate memory"),
+        ("The downstream pool failed to allocated buffer."));
+    return -1;
+  }
+map_failed:
+  {
+    GST_ELEMENT_ERROR (ffmpegdec, RESOURCE, OPEN_READ_WRITE,
+        ("Cannot access memory for read and write operation."),
+        ("The video memory allocated from downstream pool could not mapped for"
+            "read and write."));
+    return -1;
+  }
 duplicate_frame:
   {
     GST_WARNING_OBJECT (ffmpegdec, "already alloc'ed output buffer for frame");
@@ -1227,7 +1225,7 @@ get_output_buffer (GstFFMpegVidDec * ffmpegdec, GstVideoCodecFrame * frame)
   info = &ffmpegdec->output_state->info;
   if (!gst_video_frame_map (&vframe, info, frame->output_buffer,
           GST_MAP_READ | GST_MAP_WRITE))
-    goto alloc_failed;
+    goto map_failed;
 
   for (c = 0; c < AV_NUM_DATA_POINTERS; c++) {
     if (c < GST_VIDEO_INFO_N_PLANES (info)) {
@@ -1255,7 +1253,17 @@ get_output_buffer (GstFFMpegVidDec * ffmpegdec, GstVideoCodecFrame * frame)
   /* special cases */
 alloc_failed:
   {
-    GST_DEBUG_OBJECT (ffmpegdec, "allocation failed");
+    GST_ELEMENT_ERROR (ffmpegdec, RESOURCE, FAILED,
+        ("Unable to allocate memory"),
+        ("The downstream pool failed to allocated buffer."));
+    return ret;
+  }
+map_failed:
+  {
+    GST_ELEMENT_ERROR (ffmpegdec, RESOURCE, OPEN_READ_WRITE,
+        ("Cannot access memory for read and write operation."),
+        ("The video memory allocated from downstream pool could not mapped for"
+            "read and write."));
     return ret;
   }
 not_negotiated:
@@ -1408,6 +1416,16 @@ gst_ffmpegviddec_video_frame (GstFFMpegVidDec * ffmpegdec,
     *ret = get_output_buffer (ffmpegdec, out_frame);
     gst_buffer_unref (tmp);
   }
+#ifndef G_DISABLE_ASSERT
+  else {
+    GstVideoMeta *vmeta = gst_buffer_get_video_meta (out_frame->output_buffer);
+    if (vmeta) {
+      GstVideoInfo *info = &ffmpegdec->output_state->info;
+      g_assert (vmeta->width == GST_VIDEO_INFO_WIDTH (info));
+      g_assert (vmeta->height == GST_VIDEO_INFO_HEIGHT (info));
+    }
+  }
+#endif
   gst_object_unref (pool);
 
   if (G_UNLIKELY (*ret != GST_FLOW_OK))
@@ -1841,7 +1859,7 @@ gst_ffmpegviddec_decide_allocation (GstVideoDecoder * decoder, GstQuery * query)
         config_copy);
 
     /* FIXME validate and retry */
-    if (gst_buffer_pool_set_config (pool, gst_structure_copy (config_copy))) {
+    if (gst_buffer_pool_set_config (pool, config_copy)) {
       GstFlowReturn ret;
       GstBuffer *tmp;
 
@@ -1873,7 +1891,9 @@ gst_ffmpegviddec_decide_allocation (GstVideoDecoder * decoder, GstQuery * query)
     }
   }
 
-  if (have_videometa && ffmpegdec->internal_pool) {
+  if (have_videometa && ffmpegdec->internal_pool
+      && ffmpegdec->pool_width == state->info.width
+      && ffmpegdec->pool_height == state->info.height) {
     update_pool = TRUE;
     gst_object_unref (pool);
     pool = gst_object_ref (ffmpegdec->internal_pool);
